#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <malloc.h>
#include <limits.h>
#include <math.h>

#include "macros.h"
#include "misc.h"
#include "mlex.h"
#include "common.h"
#include "symbol.h"
#include "keyword.h"
#include "xmalloc.h"

#define warn_msg(fmt, arg...) \
do { \
	message(0, file_name, line, fmt, ##arg); \
} while (0)

#define err_msg(code, fmt, arg...)\
do { \
	message(1, file_name, line, fmt, ##arg); \
	exit(code); \
} while (0)

/* Lexer token data accessible outside. */
struct mlex mlex;

/* Internal variables describing lexer state. */
static char	*begin, *end;   /* Pointers to begin and end of mapped file. */
static char	*curpos;        /* Сurrent position in the mapped file. */
static char	*file_name;     /* Current file name. */
static unsigned  line;          /* Current line number. */
static char	**files;	/* Preprocessor parsed files */
static unsigned  files_size;
static unsigned  files_used;

unsigned int mlex_line_number()
{
	return line;
}

const char *mlex_file_name()
{
	return_val_if_fail(file_name != NULL, NULL);

	return file_name;
}

void mlex_file_setup(char *_begin, char *_end)
{
	return_if_fail(_begin);
	return_if_fail(_end);
	return_if_fail(_begin < _end);

	begin = _begin;
	curpos = _begin;
	end = _end;
}

void mlex_file_finalize()
{
	unsigned int i;
	/* Attempt to release resources if any left. */
	switch (mlex.token) {
	case TOKEN_ID:
		if (mlex.id != NULL)
			FREE(mlex.id);
		break;
	case TOKEN_STRING:
		if (mlex.str != NULL)
			FREE(mlex.str);
		break;
	default:
		break;
	}

	begin = end = curpos = NULL;
	line = 0;
	if (files != NULL) {
		for (i = 0; i < files_used; i++)
			xfree(files[i]);
		xfree(files);
		files = NULL;
	}
}

/* Maximum number of digits in character constant. */
#define CHAR_OCTAL_MAX          3
#define CHAR_HEX_MAX            2

/* Flags used to track state when facing with floating point or integers. */
#define MATCH_DOT               0x1
#define MATCH_EXPDIGS           0x2
#define FPT_OVERFLOW            0x4
#define FPT_UNDERFLOW           0x8
#define FPT_FROMDOUBLE          0x10
#define FPT_SUBNORMAL           0x20

static void skip_blanks(char **cur)
{
	char *pos;

	pos = *cur;
	while (isblank(*pos))
		++pos;
	*cur = pos;
}

static void put_file_name(char *file_name)
{
	if (files_size == files_used) {
		files_size += 8;
		files = xrealloc(files, sizeof(char *)*files_size);
	}
	files[files_used++] = file_name;
}

/*
 * Returns pointer to the first occurence of character c in string
 * str or NULL if LF (linefeed) or terminating zero reached.
 */
static char *strchrlf(const char *str, int c)
{
	if (str == NULL)
		return NULL;

	while (*str != '\0' && *str != c && *str != '\n')
		++str;

	if (*str != c)
		return NULL;

	return (char *) str;
}

token_t mlex_get_next_token()
{
	char *pos;
	char *pos0, *pos1;
	char *ss;
	char *cbuf;
	size_t clen;
	size_t cused;
	int base;
	int power;
	int power_temp;
	double mult;
	double number;
	int number_flags;
	unsigned long cutoff;
	int rest;
	long acc;
	int c, throwaway;
	int neg;
	int len;
	char *chr;

	cbuf = NULL;
	pos = curpos;

	/* Release allocated memory related to previous token. */
	switch (mlex.token) {
	case TOKEN_ID:
		if (mlex.id != NULL)
			FREE(mlex.id);
		break;
	case TOKEN_STRING:
		if (mlex.str != NULL)
			FREE(mlex.str);
		break;
	default:
		/* Shut up gcc complains. */
		break;
	}

	memset(&mlex, 0, sizeof(struct mlex));

	while (pos < end) {
		if (*pos == '\n')
			line++;
		if (iscntrl(*pos) && !isspace(*pos)) {
			if (*pos == 0) {
				warn_msg("ignoring NULL character");
				++pos;
				continue;
			} else {
				warn_msg("stray character code = %d", *pos);
				++pos;
				continue;
			}
		} else if (isspace(*pos)) {
			++pos;
			continue;
		} else if (*pos == '#') {
			++pos;
			skip_blanks(&pos);
			line = strtoul(pos, &pos, 10);
			skip_blanks(&pos);
			if (*pos != '\"') {
				warn_msg("a wrong format of meta information");
				continue;
			}
			++pos; /* " */
			chr = strchr(pos, '\"');
			if (chr == NULL) {
				warn_msg("a wrong format of meta information");
				continue;
			}
			*chr = 0;
			file_name = xstrdup(pos);
			put_file_name(file_name);
			*chr = '\"';
			pos = chr + 1;
			skip_blanks(&pos);
			/* Skip flag. */
			strtoul(pos, &pos, 10);
			/*
			 * Skip terminating newline character in the metainfo
			 * line. This newline must not increment line number.
			 */
			skip_blanks(&pos);
			if (*pos == '\n')
				++pos;
			continue;
		}
		break;
	}

	switch (*pos) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		/*
		 * Save position before processing float.
		 * We fall back to it if no decimal dot nor
		 * exponent digits were present in input.
		 */
do_number:
		ss = pos;
		neg = 0;
		switch (*pos) {
		case '-':
			neg = 1;
		case '+':
			++pos;
			break;
		}
		//pos0 = NULL;
		number_flags = 0;
		number = 0.0;
		power = power_temp = 0;
		while (isdigit(*pos)) {
			number = number * 10;
			number += *pos - '0';
			++pos;
		}

		if (*pos == '.') {
			pos0 = ++pos;
			while (isdigit(*pos)) {
				number = number * 10;
				number += (*pos - '0');
				++pos;
			}
			number_flags |= MATCH_DOT;
			power_temp -= (pos - pos0);
		}

		if (*pos == 'e' || *pos == 'E') {
			int exp_neg;
			exp_neg = 0;
			pos0 = pos;

			switch (*++pos) {
			case '-':
				exp_neg = 1;
			case '+':
				++pos;
			}

			pos1 = pos;
			while (isdigit(*pos)) {
				power = power * 10;
				power += *pos - '0';
				++pos;
			}

			if (exp_neg)
				power = -power;

			power += power_temp;

			if (pos == pos1) {
				/* Fall back to 'e' or 'E' if no digits. */
				pos = pos0;
			} else {
				number_flags |= MATCH_EXPDIGS;
			}
		}
		
		if (neg)
			number = -number;

		if (number == 0.0)
			goto float_done;
		
		if (number_flags & MATCH_EXPDIGS)
			power_temp = power < 0 ? -power : power;
		else {
			power = power_temp;	
			power_temp = power < 0 ? -power : power;
		}
			
		mult = 10.0;
		while (power_temp) {
			if (power_temp & 0x1) {
				if (power < 0) {
					number /= mult;
				} else {
					number *= mult;
				}
			}
			mult = mult * mult;
			power_temp >>= 1;
		}

		/* Were the decimal dot or exponent digits present? */
float_done:	if ((number_flags & (MATCH_EXPDIGS|MATCH_DOT)) == 0)
			goto do_integer;

		number_flags = 0;
		mlex.real = (float) number;

		if (isfinite(number) && !isfinite(mlex.real)) {
			number_flags |= (FPT_OVERFLOW|FPT_FROMDOUBLE);
		} else if ((number != 0.0) && (mlex.real) == 0.0) {
			number_flags |= (FPT_UNDERFLOW|FPT_FROMDOUBLE);
		} else {
			switch (fpclassify(mlex.real)) {
			case FP_INFINITE:
				number_flags |= FPT_OVERFLOW;
				break;
			case FP_SUBNORMAL:
				number_flags |= FPT_SUBNORMAL;
				break;
			}
		}

		if (number_flags) {
			if (number_flags & FPT_FROMDOUBLE) {
				warn_msg("floating point number %s "
					"when converted from double",
					(number_flags & FPT_OVERFLOW) ?
					"overflowed": "underflowed");
			} else {
				if (number_flags & FPT_OVERFLOW)
					warn_msg("floating point number "
						"overflowed will use INFINITY");
				else if (number_flags & FPT_SUBNORMAL)
					warn_msg("floating point number "
						"denormalized");
			}
		}
		curpos = pos;
		mlex.token = TOKEN_REAL;
		return TOKEN_REAL;
do_integer:
		pos = ss;
		neg = 0;
		switch (*pos) {
		case '-':
			neg = 1;
		case '+':
			++pos;
		}

		base = 8;
		if (*pos == '0') {
			if ((pos[1] == 'x') || (pos[1] == 'X')) {
				base <<= 1;
				pos0 = ++pos; /* Preserve 'X' or 'x' position. */
				++pos;
			}
		} else {
			base += 2;
		}

		acc = 0;
		throwaway = 0;
		cutoff = INT_MAX / base;
		rest = INT_MAX % base;

		while (isxdigit(*pos)) {

			if ((*pos >= '0') && (*pos <= '9'))
				c = *pos - '0';
			else if ((*pos >= 'a') && (*pos <= 'f'))
				c = *pos - 'a' + 10;
			else if ((*pos >= 'A') && (*pos <= 'F'))
				c = *pos - 'A' + 10;
			else
				break;


			if (c >= base)
				break;

			if (!throwaway) {
				if ((acc > cutoff) ||
				    ((acc == cutoff) &&
				     (c > (neg ? rest+1: rest)))) {
					throwaway = 1;
				} else {
					acc *= base;
					acc += c;
				}
			}

			++pos;
		}

		if (throwaway) {
			warn_msg("integer constant overflows "
				"will use INT_MIN or INT_MAX");
			acc = neg ? INT_MIN : INT_MAX;
		}

		if ((base == 16) && (pos == pos0)) {
			/* Return to 'x' or 'X' if no digits were present. */
			curpos = pos0;
		} else {
			curpos = pos;
		}
		mlex.integer = neg ? -((int)acc) : acc;
		mlex.token = TOKEN_INTEGER;
		return TOKEN_INTEGER;
	case '+':
		/* Handle floating point constant with preceding sign "+.123". */
		if (isdigit(pos[1]) || (pos[1] == '.' && isdigit(pos[2])))
			goto do_number;
		curpos = ++pos;
		mlex.token = TOKEN_PLUS;
		return TOKEN_PLUS;
	case '-':
		/* Handle floating point constant with preceding sign "-.123". */
		if (isdigit(pos[1]) || (pos[1] == '.' && isdigit(pos[2])))
			goto do_number;
		curpos = ++pos;
		mlex.token = TOKEN_MINUS;
		return TOKEN_MINUS;
	case '*':
		curpos = ++pos;
		mlex.token = TOKEN_ASTERISK;
		return TOKEN_ASTERISK;
	case '/':
		curpos = ++pos;
		mlex.token = TOKEN_SLASH;
		return TOKEN_SLASH;
	case '%':
		curpos = ++pos;
		mlex.token = TOKEN_PERCENT;
		return TOKEN_PERCENT;
	case '$':
		curpos = ++pos;
		mlex.token = TOKEN_DOLLAR;
		return TOKEN_DOLLAR;
	case '.':
		if (isdigit(pos[1]))
			goto do_number;
		curpos = ++pos;
		mlex.token = TOKEN_DOT;
		return TOKEN_DOT;
	case '(':
		curpos = ++pos;
		mlex.token = TOKEN_LPARENTH;
		return TOKEN_LPARENTH;
	case ')':
		curpos = ++pos;
		mlex.token = TOKEN_RPARENTH;
		return TOKEN_RPARENTH;
	case '[':
		curpos = ++pos;
		mlex.token = TOKEN_LBRACKET;
		return TOKEN_LBRACKET;
	case ']':
		curpos = ++pos;
		mlex.token = TOKEN_RBRACKET;
		return TOKEN_RBRACKET;
	case '{':
		curpos = ++pos;
		mlex.token = TOKEN_LBRACE;
		return TOKEN_LBRACE;
	case '}':
		curpos = ++pos;
		mlex.token = TOKEN_RBRACE;
		return TOKEN_RBRACE;
	case '=':
		curpos = ++pos;
		if (*pos == '=') {
			curpos = ++pos;
			mlex.token = TOKEN_EQ;
			return TOKEN_EQ;
		}
		mlex.token = TOKEN_EQUALITY;
		return TOKEN_EQUALITY;
	case ';':
		curpos = ++pos;
		mlex.token = TOKEN_SEMICOLON;
		return TOKEN_SEMICOLON;
	case ',':
		curpos = ++pos;
		mlex.token = TOKEN_COMMA;
		return TOKEN_COMMA;
	case '<':
		curpos = ++pos;
		if (*pos == '=') {
			curpos = ++pos;
			mlex.token = TOKEN_LE;
			return TOKEN_LE;
		}
		mlex.token = TOKEN_LT;
		return TOKEN_LT;
	case '>':
		curpos = ++pos;
		if (*pos == '=') {
			curpos = ++pos;
			mlex.token = TOKEN_GE;
			return TOKEN_GE;
		}
		mlex.token = TOKEN_GT;
		return TOKEN_GT;
	case '!':
		curpos = ++pos;
		if (*pos == '=') {
			curpos = ++pos;
			mlex.token = TOKEN_NE;
			return TOKEN_NE;
		}
		mlex.token = TOKEN_NOT;
		return TOKEN_NOT;
	case '\'':
		curpos = ++pos;
		chr = strchrlf(pos, '\'');

		if (chr == NULL) {
			warn_msg("missing terminating ' character");
			mlex.token = TOKEN_UNKNOWN;
			return TOKEN_UNKNOWN;
		}

		len = chr - pos;

		if (len == 0) {
			warn_msg("empty character constant");
			mlex.token = TOKEN_UNKNOWN;
			return TOKEN_UNKNOWN;
		}

		if (len == 1) {
			mlex.ch = *pos;
			curpos = ++chr;
			mlex.token = TOKEN_CHAR;
			return TOKEN_CHAR;
		} else if (*pos == '\\' && len == 2) {
			++pos;
			switch (*pos) {
			case '0':
				mlex.ch = '\0';
				break;
			case 'a':
				mlex.ch = '\a';
				break;
			case 'b':
				mlex.ch = '\b';
				break;
			case 'f':
				mlex.ch = '\f';
				break;
			case 'n':
				mlex.ch = '\n';
				break;
			case 'r':
				mlex.ch = '\r';
				break;
			case 't':
				mlex.ch = '\t';
				break;
			case 'v':
				mlex.ch = '\v';
				break;
			case '\'':
				mlex.ch = '\'';
				break;
			case '\"':
				mlex.ch = '\"';
				break;
			case '\\':
				mlex.ch = '\\';
				break;
			case '\?':
				mlex.ch = '\?';
				break;
			default:
				warn_msg("unknown escape character sequence");
				curpos = pos + len;
				mlex.token = TOKEN_UNKNOWN;
				return TOKEN_UNKNOWN;
			}
			curpos = pos + len;
			mlex.token = TOKEN_CHAR;
			return TOKEN_CHAR;
		} else if (*pos == '\\' && len >= 3) {

			unsigned int n, maxdig;

			/* Save position to calculate curpos later. */
			pos0 = pos;
			++pos;
			base = 8;

			if (pos[0] == '0') {
				if (pos[1] == 'x' || pos[1] == 'X') {
					base <<= 1;
					pos += 2;
				} else
					++pos;
			}

			switch (base) {
			case 8:
				maxdig = CHAR_OCTAL_MAX;
				break;
			case 16:
				maxdig = CHAR_HEX_MAX;
				break;
			}

			acc = n = 0;
			throwaway = 0;

			while (pos != chr) {

				if (*pos >= '0' && *pos <= '9')
					c = *pos - '0';
				else if (*pos >= 'a' && *pos <= 'f')
					c = 10 + *pos - 'a';
				else if (*pos >= 'A' && *pos <= 'F')
					c = 10 + *pos - 'A';
				else {
					throwaway = 1;
					break;
				}

				if (c < base) {
					acc *= base;
					acc += c;
				} else {
					throwaway = 1;
				}

				++pos;
				++n;
			}

			if (n > maxdig)
				throwaway = 1;

			/* When all digits were accumulated, pos must
			   be equal to the terminating quote position. */
			if (pos != chr || throwaway) {
				warn_msg("overflow in character constant conversion");
				mlex.ch = acc & 0xff;
			} else {
				mlex.ch = acc;
			}

			curpos = pos0 + len + 1;
			mlex.token = TOKEN_CHAR;
			return TOKEN_CHAR;
		}
		warn_msg("multiple characters in character constant");
		/* When multiple characters detected in character constant,
		   use the last character, the one before the terminating quote. */
		mlex.ch = pos[len-1];
		curpos = pos + len + 1;
		mlex.token = TOKEN_CHAR;
		return TOKEN_CHAR;
	case '|':
		curpos = ++pos;
		if (*pos != '|') {
			mlex.token = TOKEN_UNKNOWN;
			return TOKEN_UNKNOWN;
		}
		curpos = ++pos;
		mlex.token = TOKEN_OR;
		return TOKEN_OR;
	case '&':
		curpos = ++pos;
		if (*pos != '&') {
			mlex.token = TOKEN_AMPERSAND;
			return TOKEN_AMPERSAND;
		}
		curpos = ++pos;
		mlex.token = TOKEN_AND;
		return TOKEN_AND;
	case '\"':
		curpos = ++pos;
		chr = pos;
		/* Locate pairing double quote. */
		while (chr[0] != '\0' && chr != end) {
			/* Detect the case when a pairing quote
			   is not present at the current line. */
			if (chr[0] == '\n')
				break;

			if (chr[0] == '"') {
				mlex.str = xstrndup(pos, chr - pos);
				curpos = ++chr;
				mlex.token = TOKEN_STRING;
				return TOKEN_STRING;
			}

			if (chr[0] == '\\') {
				if (chr[1] == '"') {
					chr += 2;
					continue;
				}
			}
			++chr;
		}
		warn_msg("missing pairing \" character in string constant");
		curpos = chr;
		mlex.token = TOKEN_UNKNOWN;
		return TOKEN_UNKNOWN;
	default:
		if (pos == end) {
			mlex.token = TOKEN_EOF;
			return TOKEN_EOF;
		}
		/* Try to locate id or symbol. */
		if ((*pos == '_') || isalpha(*pos)) {
			struct keyword *kword;
			struct symbol *symbol;
			clen = cused = 0;
			do {
				if ((cused+1) >= clen) {
					clen += 64;
					cbuf = xrealloc(cbuf, clen);
				}
				cbuf[cused++] = *pos;
				++pos;
			} while ((*pos == '_') || isalnum(*pos));			
			cbuf[cused++] = 0;
			curpos = pos;
			/* check string for keyword */
			kword = keyword_table_lookup(cbuf);
			if (kword != NULL) {
				mlex.token = kword->type;
				FREE(cbuf);
				return kword->type;
			}
			/* Try to find string in symbol tables. */
			symbol = symbol_table_lookup_all(cbuf);
			if (symbol != NULL) {
				mlex.id = xstrdup(symbol->name);
			} else {
				mlex.id = xstrdup(cbuf);
			}
			FREE(cbuf);
			mlex.token = TOKEN_ID;
			return TOKEN_ID;
		}
		break;
	}
	curpos = ++pos;
	mlex.token = TOKEN_UNKNOWN;
	return TOKEN_UNKNOWN;
}

